import random
from typing import Tuple, List
from tekleo_common_message_protocol import OdSample, OdLabeledItem, PointRelative
from tekleo_common_utils import UtilsImage, UtilsOpencv
from tekleo_common_utils_ai.dataset_modification.abstract_dataset_modifier import AbstractDatasetModifier
from injectable import injectable, autowired, Autowired


@injectable
class DatasetModifierBorder(AbstractDatasetModifier):
    @autowired
    def __init__(self,
                 min_border_ratio: float, max_border_ratio: float, border_type: str, border_colors: List[Tuple[int, int, int]], random_seed: int,
                 utils_image: Autowired(UtilsImage), utils_opencv: Autowired(UtilsOpencv),
                 ):
        self.utils_image = utils_image
        self.utils_opencv = utils_opencv
        self.min_border_ratio = min_border_ratio
        self.max_border_ratio = max_border_ratio
        self.border_type = border_type
        self.border_colors = border_colors
        self.random_seed = random_seed
        self.random = random.Random()
        self.random.seed(self.random_seed)


    def apply(self, sample: OdSample) -> OdSample:
        # Convert image to opencv
        image_pil = sample.image
        image_cv = self.utils_image.convert_image_pil_to_image_cv(image_pil)
        image_width, image_height = self.utils_opencv.get_dimensions_wh(image_cv)

        # Determine border color and ratio
        border_color = self.random.choice(self.border_colors)
        border_ratio_x = self.random.uniform(self.min_border_ratio, self.max_border_ratio)
        border_ratio_y = self.random.uniform(self.min_border_ratio, self.max_border_ratio)

        # Find pixel border size
        border_size_x = int(image_width * border_ratio_x)
        border_size_y = int(image_height * border_ratio_y)

        # Adjust border size according to type
        if self.border_type == "x":
            border_size_y = 0
        elif self.border_type == "y":
            border_size_x = 0
        elif self.border_type == "both":
            pass

        # Apply border to the image
        image_cv = self.utils_opencv.border(image_cv, border_size_y, border_size_y, border_size_x, border_size_x, border_color)
        new_image_width, new_image_height = self.utils_opencv.get_dimensions_wh(image_cv)

        # Convert back to pil
        image_pil = self.utils_image.convert_image_cv_to_image_pil(image_cv)

        # Move all mask points
        new_items = []
        for item in sample.items:
            new_mask = []
            for point in item.mask:
                old_pixel_x = int(point.x * image_width)
                old_pixel_y = int(point.y * image_height)
                moved_pixel_x = old_pixel_x + 1 * border_size_x
                moved_pixel_y = old_pixel_y + 1 * border_size_y
                new_point = PointRelative(moved_pixel_x / new_image_width, moved_pixel_y / new_image_height)
                new_mask.append(new_point)
            new_items.append(OdLabeledItem(
                item.label,
                new_mask
            ))

        # Generate new name
        new_name = sample.name
        if "_mod_" in new_name:
            if "_border_" in new_name:
                new_name = new_name + "_" + self.border_type
            else:
                new_name = new_name + "_border_" + self.border_type
        else:
            new_name = sample.name + "_mod_border_" + self.border_type

        # Return new sample
        return OdSample(
            new_name,
            image_pil,
            new_items
        )



import os, warnings
warnings.simplefilter("ignore")
os.environ["PYTHONWARNINGS"] = "ignore"

from .utils_dataset_coco import UtilsDatasetCoco
from .utils_dataset_google import UtilsDatasetGoogle
from .utils_dataset_labelme import UtilsDatasetLabelme
from .utils_dataset_pascalvoc import UtilsDatasetPascalvoc
from .utils_detectron_model import UtilsDetectronModel
from .utils_visualize_od import UtilsVisualizeOd
from .dataset_modification.dataset_modification_pipe import DatasetModificationPipe, BehaviorRandom, BehaviorChaining, BehaviorOriginals
from .dataset_modification.dataset_modifier_blur import DatasetModifierBlur
from .dataset_modification.dataset_modifier_border import DatasetModifierBorder
from .dataset_modification.dataset_modifier_brightness import DatasetModifierBrightness
from .dataset_modification.dataset_modifier_contrast import DatasetModifierContrast
from .dataset_modification.dataset_modifier_crop import DatasetModifierCrop
from .dataset_modification.dataset_modifier_hue import DatasetModifierHue
from .dataset_modification.dataset_modifier_saturation import DatasetModifierSaturation
from .dataset_modification.dataset_modifier_sharpen import DatasetModifierSharpen

__all__ = [
    UtilsDatasetCoco,
    UtilsDatasetGoogle,
    UtilsDatasetLabelme,
    UtilsDatasetPascalvoc,
    UtilsDetectronModel,
    UtilsVisualizeOd,
    DatasetModificationPipe, BehaviorRandom, BehaviorChaining, BehaviorOriginals,
    DatasetModifierBlur, DatasetModifierBorder, DatasetModifierBrightness,
    DatasetModifierContrast, DatasetModifierCrop, DatasetModifierHue,
    DatasetModifierSaturation, DatasetModifierSharpen,
]



import random
from tekleo_common_message_protocol import OdSample
from tekleo_common_utils import UtilsImage, UtilsOpencv
from tekleo_common_utils_ai.dataset_modification.abstract_dataset_modifier import AbstractDatasetModifier
from injectable import injectable, autowired, Autowired


@injectable
class DatasetModifierSaturation(AbstractDatasetModifier):
    @autowired
    def __init__(
            self, min_saturation_ratio: float, max_saturation_ratio: float, saturation_application: str, random_seed: int,
            utils_image: Autowired(UtilsImage), utils_opencv: Autowired(UtilsOpencv),
    ):
        self.utils_image = utils_image
        self.utils_opencv = utils_opencv
        self.min_saturation_ratio = min_saturation_ratio
        self.max_saturation_ratio = max_saturation_ratio
        self.saturation_application = saturation_application
        self.random_seed = random_seed

        # Instance of random generator
        self.random = random.Random()
        self.random.seed(self.random_seed)

    def apply(self, sample: OdSample) -> OdSample:
        # Convert image to opencv
        image_pil = sample.image
        image_cv = self.utils_image.convert_image_pil_to_image_cv(image_pil)

        # Determine
        saturation_ratio = self.random.uniform(self.min_saturation_ratio, self.max_saturation_ratio)

        # Determine the sign
        saturation_sign = 1
        if self.saturation_application == 'decrease':
            saturation_sign = -1
        elif self.saturation_application == 'increase':
            saturation_sign = 1
        elif self.saturation_application == 'both':
            should_increase = self.random.choice([True, False])
            if should_increase:
                saturation_sign = 1
            else:
                saturation_sign = - 1

        # Convert back to 127
        saturation_delta = saturation_ratio
        saturation_value = 1 + saturation_sign * saturation_delta
        # print("saturation_value=" + str(saturation_value))

        # Apply saturation to the image
        image_cv = self.utils_opencv.saturation(image_cv, saturation_coefficient=saturation_value)

        # Convert back to pil
        image_pil = self.utils_image.convert_image_cv_to_image_pil(image_cv)

        # Generate new name
        new_name = sample.name
        if "_mod_" in new_name:
            new_name = new_name + "_saturation_" + self.saturation_application[0:4]
        else:
            new_name = sample.name + "_mod_saturation_" + self.saturation_application[0:4]

        # Return new sample
        return OdSample(
            new_name,
            image_pil,
            sample.items
        )



from typing import List
import random
from tekleo_common_message_protocol import OdSample
from tekleo_common_utils_ai.dataset_modification.abstract_dataset_modifier import AbstractDatasetModifier


class BehaviorRandom:
    ALL = 'all' # apply all modifiers
    RANDOM = 'random' # apply random modifiers


class BehaviorChaining:
    CHAINED = 'chained' # chained to produce a single image
    UNIQUE = 'unique' # each modifier produces a new image


class BehaviorOriginals:
    KEEP = 'keep' # keep the originals
    OVERWRITE = 'overwrite' # overwrite the originals


class DatasetModificationPipe:
    # behavior_dataset_ratio_to_process -> 1.0 work on all images | 0.15 work on 15% of images
    # behavior_random -> "all" apply all modifiers | "random" apply random modifiers
    # behavior_chaining -> "chained" chained to produce a single image | "unique" each modifier produces a new image
    # behavior_originals -> "keep" keep the originals | "overwrite" overwrite the originals
    def __init__(
            self, modifiers: List[AbstractDatasetModifier],
            behavior_dataset_ratio_to_process: float, behavior_random_seed: int,
            behavior_random: BehaviorRandom, behavior_chaining: BehaviorChaining, behavior_originals: BehaviorOriginals
    ):
        self.modifiers = modifiers
        self.behavior_dataset_ratio_to_process = behavior_dataset_ratio_to_process
        self.behavior_random_seed = behavior_random_seed
        self.behavior_random = behavior_random # apply all modifiers | apply random modifiers
        self.behavior_chaining = behavior_chaining # chained to produce a single image | each modifier produces a new image
        self.behavior_originals = behavior_originals # keep the originals | overwrite the originals

        # Instance of random generator
        self.random = random.Random()
        self.random.seed(self.behavior_random_seed)

    def process(self, samples: List[OdSample]) -> List[OdSample]:
        new_samples = []

        # Determine which samples to process
        samples_to_process = samples.copy()
        if self.behavior_dataset_ratio_to_process != 1.0:
            num_of_samples_to_process = int(self.behavior_dataset_ratio_to_process * len(samples))
            samples_to_process = self.random.shuffle(samples_to_process)
            samples_to_process = samples_to_process[0:num_of_samples_to_process]

        for sample in samples_to_process:
            # Determine which modifiers to apply to the sample
            modifiers_to_apply = []
            for modifier in self.modifiers:
                if self.behavior_random == BehaviorRandom.RANDOM:
                    should_apply = self.random.choice([True, False])
                    if should_apply:
                        modifiers_to_apply.append(modifier)
                elif self.behavior_random == BehaviorRandom.ALL:
                    modifiers_to_apply.append(modifier)

            # Keep modifiers chained
            if self.behavior_chaining == BehaviorChaining.CHAINED:
                new_sample = sample
                for modifier in modifiers_to_apply:
                    new_sample = modifier.apply(new_sample)
                new_samples.append(new_sample)

            # Produce unique image for each modifier
            elif self.behavior_chaining == BehaviorChaining.UNIQUE:
                for modifier in modifiers_to_apply:
                    new_sample = modifier.apply(sample)
                    new_samples.append(new_sample)

        # Add originals to the dataset
        if self.behavior_originals == BehaviorOriginals.KEEP:
            new_samples.extend(samples)

        return new_samples




import random
from tekleo_common_message_protocol import OdSample, OdLabeledItem, PointRelative
from tekleo_common_utils import UtilsImage, UtilsOpencv
from tekleo_common_utils_ai.dataset_modification.abstract_dataset_modifier import AbstractDatasetModifier
from injectable import injectable, autowired, Autowired


@injectable
class DatasetModifierCrop(AbstractDatasetModifier):
    @autowired
    def __init__(self,
                 min_crop_ratio: float, max_crop_ratio: float, crop_type: str, random_seed: int,
                 utils_image: Autowired(UtilsImage), utils_opencv: Autowired(UtilsOpencv),
                 ):
        self.utils_image = utils_image
        self.utils_opencv = utils_opencv
        self.min_crop_ratio = min_crop_ratio
        self.max_crop_ratio = max_crop_ratio
        self.crop_type = crop_type
        self.random_seed = random_seed
        self.random = random.Random()
        self.random.seed(self.random_seed)


    def apply(self, sample: OdSample) -> OdSample:
        # Convert image to opencv
        image_pil = sample.image
        image_cv = self.utils_image.convert_image_pil_to_image_cv(image_pil)
        image_width, image_height = self.utils_opencv.get_dimensions_wh(image_cv)

        # Determine crop ratio
        crop_ratio_x = self.random.uniform(self.min_crop_ratio, self.max_crop_ratio)
        crop_ratio_y = self.random.uniform(self.min_crop_ratio, self.max_crop_ratio)

        # Find pixel crop size
        crop_size_x = int(image_width * crop_ratio_x)
        crop_size_y = int(image_height * crop_ratio_y)

        # Adjust crop size according to type
        if self.crop_type == "x":
            crop_size_y = 0
        elif self.crop_type == "y":
            crop_size_x = 0
        elif self.crop_type == "both":
            pass

        # Find mask coordinates
        masks_points_x = []
        masks_points_y = []
        for item in sample.items:
            for point in item.mask:
                masks_points_x.append(int(point.x * image_width))
                masks_points_y.append(int(point.y * image_height))
        masks_points_x = sorted(list(set(masks_points_x)))
        masks_points_y = sorted(list(set(masks_points_y)))

        # Fins max allowed crop
        allowed_crop_l = min(masks_points_x)
        allowed_crop_r = max(masks_points_x)
        allowed_crop_t = min(masks_points_y)
        allowed_crop_b = max(masks_points_y)
        if allowed_crop_t > 5:
            allowed_crop_t = allowed_crop_t - 4
        if allowed_crop_l > 5:
            allowed_crop_l = allowed_crop_l - 4
        if allowed_crop_r + 5 < image_width:
            allowed_crop_r = allowed_crop_r + 4
        if allowed_crop_b + 5 < image_height:
            allowed_crop_b = allowed_crop_b + 4

        # Apply crop to the image
        crop_y_start = min(crop_size_y, allowed_crop_t)
        crop_y_end = max(image_height - crop_size_y, allowed_crop_b)
        crop_x_start = min(crop_size_x, allowed_crop_l)
        crop_x_end = max(image_width - crop_size_x, allowed_crop_r)
        image_cv = image_cv[crop_y_start:crop_y_end, crop_x_start:crop_x_end]
        new_image_width, new_image_height = self.utils_opencv.get_dimensions_wh(image_cv)

        # Convert back to pil
        image_pil = self.utils_image.convert_image_cv_to_image_pil(image_cv)

        # Move all mask points
        new_items = []
        for item in sample.items:
            new_mask = []
            for point in item.mask:
                old_pixel_x = int(point.x * image_width)
                old_pixel_y = int(point.y * image_height)
                moved_pixel_x = old_pixel_x - crop_x_start
                moved_pixel_y = old_pixel_y - crop_y_start
                new_point = PointRelative(moved_pixel_x / new_image_width, moved_pixel_y / new_image_height)
                new_mask.append(new_point)
            new_items.append(OdLabeledItem(
                item.label,
                new_mask
            ))

        # Generate new name
        new_name = sample.name
        if "_mod_" in new_name:
            if "_crop_" in new_name:
                new_name = new_name + "_" + self.crop_type
            else:
                new_name = new_name + "_crop_" + self.crop_type
        else:
            new_name = sample.name + "_mod_crop_" + self.crop_type

        # Return new sample
        return OdSample(
            new_name,
            image_pil,
            new_items
        )



import random
from tekleo_common_message_protocol import OdSample
from tekleo_common_utils import UtilsImage, UtilsOpencv
from tekleo_common_utils_ai.dataset_modification.abstract_dataset_modifier import AbstractDatasetModifier
from injectable import injectable, autowired, Autowired


@injectable
class DatasetModifierBlur(AbstractDatasetModifier):
    @autowired
    def __init__(
            self, min_blur_ratio: float, max_blur_ratio: float, random_seed: int,
            utils_image: Autowired(UtilsImage), utils_opencv: Autowired(UtilsOpencv),
    ):
        self.utils_image = utils_image
        self.utils_opencv = utils_opencv
        self.min_blur_ratio = min_blur_ratio
        self.max_blur_ratio = max_blur_ratio
        self.random_seed = random_seed

        # Instance of random generator
        self.random = random.Random()
        self.random.seed(self.random_seed)

    def apply(self, sample: OdSample) -> OdSample:
        # Convert image to opencv
        image_pil = sample.image
        image_cv = self.utils_image.convert_image_pil_to_image_cv(image_pil)
        image_width, image_height = self.utils_opencv.get_dimensions_wh(image_cv)

        # Determine
        blur_ratio = self.random.uniform(self.min_blur_ratio, self.max_blur_ratio)

        # Convert back to 127
        blur_baseline = min(image_height, image_width)
        blur_x = int(blur_ratio * blur_baseline)
        blur_y = int(blur_ratio * blur_baseline)
        if blur_x % 2 == 0:
            blur_x = blur_x + 1
        if blur_y % 2 == 0:
            blur_y = blur_y + 1
        # print("blur_x=" + str(blur_x) + ", blur_y=" + str(blur_y))

        # Apply blur to the image
        image_cv = self.utils_opencv.blur_gaussian(image_cv, blur_x, blur_y)

        # Convert back to pil
        image_pil = self.utils_image.convert_image_cv_to_image_pil(image_cv)

        # Generate new name
        new_name = sample.name
        if "_mod_" in new_name:
            new_name = new_name + "_blur"
        else:
            new_name = sample.name + "_mod_blur"

        # Return new sample
        return OdSample(
            new_name,
            image_pil,
            sample.items
        )
